"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.resolvePartLinks = exports.resolveLinks = void 0;
const ts = __importStar(require("typescript"));
const models_1 = require("../../models");
const declarationReference_1 = require("./declarationReference");
const declarationReferenceResolver_1 = require("./declarationReferenceResolver");
const urlPrefix = /^(http|ftp)s?:\/\//;
function resolveLinks(comment, reflection, attemptExternalResolve) {
    comment.summary = resolvePartLinks(reflection, comment.summary, attemptExternalResolve);
    for (const tag of comment.blockTags) {
        tag.content = resolvePartLinks(reflection, tag.content, attemptExternalResolve);
    }
    if (reflection instanceof models_1.DeclarationReflection && reflection.readme) {
        reflection.readme = resolvePartLinks(reflection, reflection.readme, attemptExternalResolve);
    }
}
exports.resolveLinks = resolveLinks;
function resolvePartLinks(reflection, parts, attemptExternalResolve) {
    return parts.flatMap((part) => processPart(reflection, part, attemptExternalResolve));
}
exports.resolvePartLinks = resolvePartLinks;
function processPart(reflection, part, attemptExternalResolve) {
    if (part.kind === "inline-tag") {
        if (part.tag === "@link" ||
            part.tag === "@linkcode" ||
            part.tag === "@linkplain") {
            return resolveLinkTag(reflection, part, attemptExternalResolve);
        }
    }
    return part;
}
function resolveLinkTag(reflection, part, attemptExternalResolve) {
    let pos = 0;
    const end = part.text.length;
    while (pos < end && ts.isWhiteSpaceLike(part.text.charCodeAt(pos))) {
        pos++;
    }
    // Try to parse one
    const declRef = (0, declarationReference_1.parseDeclarationReference)(part.text, pos, end);
    let target;
    let defaultDisplayText;
    if (declRef) {
        // Got one, great! Try to resolve the link
        target = (0, declarationReferenceResolver_1.resolveDeclarationReference)(reflection, declRef[0]);
        pos = declRef[1];
        if (target) {
            defaultDisplayText = target.name;
        }
        else {
            // If we didn't find a link, it might be a @link tag to an external symbol, check that next.
            target = attemptExternalResolve(declRef[0]);
            if (target) {
                defaultDisplayText = part.text.substring(0, pos);
            }
        }
    }
    if (!target) {
        if (urlPrefix.test(part.text)) {
            const wsIndex = part.text.search(/\s/);
            target =
                wsIndex === -1 ? part.text : part.text.substring(0, wsIndex);
            pos = target.length;
            defaultDisplayText = target;
        }
    }
    // Remaining text after an optional pipe is the link text, so advance
    // until that's consumed.
    while (pos < end && ts.isWhiteSpaceLike(part.text.charCodeAt(pos))) {
        pos++;
    }
    if (pos < end && part.text[pos] === "|") {
        pos++;
    }
    if (!target) {
        return part;
    }
    part.target = target;
    part.text = part.text.substring(pos).trim() || defaultDisplayText;
    return part;
}
